; *********************************************************************************
; * Ficheiro:  lab7-cooperativas-teclado-OK.asm
; * Descrio: Este programa ilustra o funcionamento de vrias rotinas cooperativas,
; *			em que nenhuma das rotinas  bloqueante.
; *			Igual ao lab7-cooperativas-teclado-bloqueante.asm, mas em que a rotina
; *			TECLADO retorna sempre, haja ou no tecla carregada (na 4 linha do teclado).
; *
; *********************************************************************************

; *********************************************************************************
; * Constantes
; *********************************************************************************
COMANDOS				EQU	6000H			; endereo de base dos comandos do MediaCenter

DEFINE_LINHA    		EQU COMANDOS + 0AH		; endereo do comando para definir a linha
DEFINE_COLUNA   		EQU COMANDOS + 0CH		; endereo do comando para definir a coluna
ESCREVE_8_PIXELS    	EQU COMANDOS + 1CH		; endereo do comando para escrever 8 pixels
APAGA_AVISO     		EQU COMANDOS + 40H		; endereo do comando para apagar o aviso de nenhum cenrio selecionado
APAGA_ECR	 		EQU COMANDOS + 02H		; endereo do comando para apagar todos os pixels j desenhados

N_LINHAS				EQU  32				; nmero de linhas do cr
BARRA				EQU  0FFH				; valor do byte usado para representar a barra

DISPLAYS				EQU  0A000H			; endereo do perifrico que liga aos displays
DELAY				EQU  100H				; valor usado para implementar um atraso temporal    

TEC_LIN				EQU  0C000H			; endereo das linhas do teclado (perifrico POUT-2)
TEC_COL				EQU  0E000H			; endereo das colunas do teclado (perifrico PIN)
LINHA				EQU  8				; linha a testar (4 linha, 1000b)
MASCARA				EQU	0FH				; para isolar os 4 bits de menor peso, ao ler as colunas do teclado

; *********************************************************************************
; * Dados 
; *********************************************************************************
	PLACE       1000H
pilha:
	STACK 100H			; espao reservado para a pilha 
						; (200H bytes, pois so 100H words)
SP_inicial:				; este  o endereo (1200H) com que o SP deve ser 
						; inicializado. O 1. end. de retorno ser 
						; armazenado em 11FEH (1200H-2)
                              
; Tabela das rotinas de interrupo
tab:
	WORD rot_int_0			; rotina de atendimento da interrupo 0
	WORD rot_int_1			; rotina de atendimento da interrupo 1
	WORD rot_int_2			; rotina de atendimento da interrupo 2
	WORD rot_int_3			; rotina de atendimento da interrupo 3

evento_int:
	WORD 0				; se 1, indica que a interrupo 0 ocorreu
	WORD 0				; se 1, indica que a interrupo 1 ocorreu
	WORD 0				; se 1, indica que a interrupo 2 ocorreu
	WORD 0				; se 1, indica que a interrupo 3 ocorreu
                              
linha_barra:
	WORD 0				; linha em que a 1 barra est
	WORD 0				; linha em que a 2 barra est
	WORD 0				; linha em que a 3 barra est
	WORD 0				; linha em que a 4 barra est

contador:
	WORD 0				; contador usado para mostrar nos displays
                              
contador_atraso:
	WORD DELAY			; contador usado para gerar o atraso
          
coluna_carregada:
	WORD 0				; varivel que indica se uma tecla foi carregada
						; 0 - nenhuma tecla carregada
						; 1, 2, 4 ou 8 - tecla carregada, e o valor indica a coluna da tecla) 
          
; *********************************************************************************
; * Cdigo
; *********************************************************************************
	PLACE   0				; o cdigo tem de comear em 0000H
inicio:
	MOV  SP, SP_inicial		; inicializa SP para a palavra a seguir
						;  ltima da pilha
                            
	MOV  BTE, tab			; inicializa BTE (registo de Base da Tabela de Excees)

	MOV  [APAGA_AVISO], R1	; apaga o aviso de nenhum cenrio selecionado (o valor de R1 no  relevante)
	MOV  [APAGA_ECR], R1	; apaga todos os pixels j desenhados (o valor de R1 no  relevante)
     
	EI0					; permite interrupes 0
	EI1					; permite interrupes 1
	EI2					; permite interrupes 2
	EI3					; permite interrupes 3
	EI					; permite interrupes (geral)

; ciclo das rotinas cooperativas no programa principal
ciclo:

	CALL teclado			; verifica se alguma tecla foi carregada

	CALL anima_displays		; faz os displays avanar de uma unidade

	; Estes 4 CALLs podiam ser feitos num ciclo, incrementando o R3, mas assim  mais explcito
	MOV  R3, 0			; n da barra
	CALL anima_barra		; faz a barra nesta coluna descer de uma linha. Se chegar ao fundo, passa ao topo

	MOV  R3, 1			; n da barra
	CALL anima_barra		; faz a barra nesta coluna descer de uma linha. Se chegar ao fundo, passa ao topo

	MOV  R3, 2			; n da barra
	CALL anima_barra		; faz a barra nesta coluna descer de uma linha. Se chegar ao fundo, passa ao topo

	MOV  R3, 3			; n da barra
	CALL anima_barra		; faz a barra nesta coluna descer de uma linha. Se chegar ao fundo, passa ao topo

	JMP  ciclo


; **********************************************************************
; Rotinas cooperativas 
; **********************************************************************


; **********************************************************************
; TECLADO - Rotina cooperativa que deteta quando se carrega numa tecla na 4 linha
;		  do teclado. J no  bloqueante e retorna logo, haja ou no uma tecla carregada.
; **********************************************************************
teclado:
	PUSH R0
	PUSH R1
	PUSH R2
	PUSH R3
	PUSH R5
	MOV  R2, TEC_LIN         ; endereo do perifrico das linhas
	MOV  R3, TEC_COL         ; endereo do perifrico das colunas
	MOV  R5, MASCARA		; para isolar os 4 bits de menor peso, ao ler as colunas do teclado

	MOV  R1, LINHA           ; testar a linha 4 
	MOVB [R2], R1            ; escrever no perifrico de sada (linhas)
	MOVB R0, [R3]            ; ler do perifrico de entrada (colunas)
	AND  R0, R5			; elimina bits para alm dos bits 0-3
	CMP  R0, 0               ; h tecla premida?
	JNZ  ha_tecla
	MOV  R2, 0               ; nenhuma tecla premida
	JMP  sai_teclado
ha_tecla:
	MOV  R2, R0              ; houve uma tecla premida
sai_teclado:
	MOV  [coluna_carregada], R2	; atualiza na varivel a informao sobre se houve ou no tecla premida
							; O uso de R2  redundante (bastava guardar R0), mas assim  mais explcito
     POP  R5
     POP  R3
     POP  R2
     POP  R1
     POP  R0
     RET                      ; retorna sempre, haja ou no uma tecla carregada


; **********************************************************************
; ANIMA_DISPLAYS - Rotina cooperativa que aumenta ou diminui o valor dos displays de uma unidade,
;			    consoante no haja, ou haja, respetivamente, uma tecla premida na 4 linha do teclado
; **********************************************************************
anima_displays:
	PUSH R0
	PUSH R1
	PUSH R2
	CALL atraso              ; espera tempo
	CMP  R1, 0               ; tempo de espera chegou ao fim?
	JNZ  sai_anima_displays
	MOV  R2, [contador]		; obtm valor do contador, cujo valor vai ser mostrado nos displays
	MOV  R0, DISPLAYS        ; endereo do perifrico que liga aos displays
	MOVB [R0], R2            ; mostra o valor do contador nos displays
	MOV  R1, [coluna_carregada]	; obtm em R1 a informao sobre se h ou no tecla premida
	CMP  R1, 0               ; h tecla premida?
	JZ  cresce
	SUB  R2, 1               ; h, diminui o valor dos displays
	JMP  atualiza
cresce:
	ADD  R2, 1               ; no, aumenta o valor dos displays
atualiza:
	MOV  [contador], R2		; atualiza valor do contador
sai_anima_displays:
	POP  R2
	POP  R1
	POP  R0
	RET

; **********************************************************************
; ANIMA_BARRA - Rotina cooperativa que desenha e faz descer uma barra de 8 pixels no ecr,
;			 numa dada coluna. Se chegar ao fundo, passa ao topo.
;			 A linha em que o byte  escrito  guardada na varivel linha_barra, que 
;			 uma tabela de quatro variveis simples, uma para cada barra
; Argumentos: R3 - N da barra (0 a 3)
; **********************************************************************
anima_barra:
	PUSH R1
	PUSH R2
	PUSH R3
	PUSH R4
	PUSH R5
	PUSH R6
	MOV  R6, R3			; cpia de R3 (para no destruir R3)
	SHL  R6, 1			; multiplica o n da barra por 2 (porque a linha_barra  uma tabela de words)

	MOV  R5, evento_int
	MOV  R2, [R5+R6]		; valor da varivel que diz se houve uma interrupo com o mesmo nmero da coluna
	CMP  R2, 0
	JZ   sai_anima_barra	; se no houve interrupo, vai-se embora
	MOV  R2, 0
	MOV  [R5+R6], R2		; coloca a zero o valor da varivel que diz se houve uma interrupo (consome evento)

	MOV  R5, linha_barra
	MOV  R2, [R5+R6]		; linha em que a barra est
	SHL  R3, 3			; multiplica o n da barra por 8 para dar a coluna onde desenhar o byte
	MOV  R1, 0			; para apagar a barra
	CALL escreve_byte		; apaga a barra do ecr
	ADD  R2, 1			; passa  linha abaixo
	MOV  R4, N_LINHAS
	CMP  R2, R4			; j estava na linha do fundo?
	JLT  escreve
	MOV  R2, 0			; volta ao topo do ecr
escreve:
	MOV  [R5+R6], R2		; atualiza na tabela a linha em que esta barra est
	MOV  R1, BARRA			; valor da barra
	CALL escreve_byte		; escreve a barra na nova linha
sai_anima_barra:
	POP  R6
	POP  R5
	POP  R4
	POP  R3
	POP  R2
	POP  R1
	RET


; **********************************************************************
; Rotinas de interrupo 
; **********************************************************************


; **********************************************************************
; ROT_INT_0 - 	Rotina de atendimento da interrupo 0
;			Assinala o evento na componente 0 da varivel evento_int
; **********************************************************************
rot_int_0:
	PUSH R0
	PUSH R1
	MOV  R0, evento_int
	MOV  R1, 1			; assinala que houve uma interrupo 0
	MOV  [R0], R1			; na componente 0 da varivel evento_int
	POP  R1
	POP  R0
	RFE

; **********************************************************************
; ROT_INT_1 - 	Rotina de atendimento da interrupo 1
;			Assinala o evento na componente 1 da varivel evento_int
; **********************************************************************
rot_int_1:
	PUSH R0
	PUSH R1
	MOV  R0, evento_int
	MOV  R1, 1			; assinala que houve uma interrupo 0
	MOV  [R0+2], R1		; na componente 1 da varivel evento_int
						; Usa-se 2 porque cada word tem 2 bytes
	POP  R1
	POP  R0
	RFE

; **********************************************************************
; ROT_INT_2 -	Rotina de atendimento da interrupo 2
;			Assinala o evento na componente 2 da varivel evento_int
; **********************************************************************
rot_int_2:
	PUSH R0
	PUSH R1
	MOV  R0, evento_int
	MOV  R1, 1			; assinala que houve uma interrupo 0
	MOV  [R0+4], R1		; na componente 2 da varivel evento_int
						; Usa-se 4 porque cada word tem 2 bytes
	POP  R1
	POP  R0
	RFE

; **********************************************************************
; ROT_INT_3 -	Rotina de atendimento da interrupo 3
;			Assinala o evento na componente 3 da varivel evento_int
; **********************************************************************
rot_int_3:
	PUSH R0
	PUSH R1
	MOV  R0, evento_int
	MOV  R1, 1			; assinala que houve uma interrupo 0
	MOV  [R0+6], R1		; na componente 3 da varivel evento_int
						; Usa-se 6 porque cada word tem 2 bytes
	POP  R1
	POP  R0
	RFE

     
; **********************************************************************
; Rotinas auxiliares 
; **********************************************************************


; **********************************************************************
; ESCREVE_BYTE - Escreve um byte no ecr, com a cor da caneta
; Argumentos:	R1 - Valor do byte a escrever
;			R2 - Linha onde escrever o <<<byte (entre 0 e N_LINHAS - 1)
;			R3 - Coluna a partir da qual a barra deve ser desenhada 
; **********************************************************************
escreve_byte:
	MOV  [DEFINE_LINHA], R2		; seleciona a linha
	MOV  [DEFINE_COLUNA], R3		; seleciona a coluna
	MOV  [ESCREVE_8_PIXELS], R1	; escreve os 8 pixels correspondentes ao byte
	RET

; **********************************************************************
; ATRASO - Faz DELAY iteraes, para implementar um atraso no tempo,
;		 de forma no bloqueante.
; Argumentos: Nenhum
; Saidas:		R1 - Se 0, o atraso chegou ao fim
; **********************************************************************
atraso:
	PUSH R2
	MOV  R1, [contador_atraso]	; obtm valor do contador do atraso
	SUB  R1, 1
	MOV  [contador_atraso], R1	; atualiza valor do contador do atraso
	JNZ  sai
	MOV  R2, DELAY
	MOV  [contador_atraso], R2	; volta a colocar o valor inicial no contador do atraso
sai:
	POP  R2
	RET

